'use strict'

// keeps a history of cpu & memory aggregated by percentile & time
const RecentCache = require('../lib/recent_cache')

const BUCKETS = 25

function Entry (time) {
  this.time = time
  // entries is a coordinate space from "0,0" to "9,9"
  // for a maximum of 100 records per entry, though in practice
  // our data is much more sparse than that
  this.gridspace = {}
}

class QuantileHistory {
  constructor () {
    // TBD abstract this?
    // 1.3 minutes (so each column is 3 seconds)
    this.timeRange = 1.3 * 60 * 1000
    // plus one slush column because time flooring
    this.timeBuckets = BUCKETS + 1
    this.quantiles = 10

    this.bucketInterval = this.timeRange / this.timeBuckets

    this.cache = new RecentCache(this.timeRange)
    this.seed(Date.now())
  }

  seed (from) {
    const current = this.floorTimestamp(from)
    for (let i = 0; i < this.timeBuckets; i++) {
      const timeBucket = current - this.bucketInterval * i
      let entry = this.cache.entries().find((e) => {
        return e.time === timeBucket
      })
      if (entry == null) {
        entry = new Entry(timeBucket)
        this.cache.insert(entry)
      }
    }
  }

  floorTimestamp (ts) {
    return ((ts / this.bucketInterval) | 0) * this.bucketInterval
  }

  toQuantile (value) {
    // range of values is 0>100
    let rounded = Math.floor(value * 10)
    if (rounded > 10) {
      rounded = 9
    }
    if (rounded < 0) {
      rounded = 0
    }
    return rounded
  }

  insert (record) {
    if (record == null) {
      return
    }
    const timeBucket = this.floorTimestamp(record.time)
    if (timeBucket < Date.now() - this.timeRange) {
      // record is too old already
      return
    }
    const cpu = this.toQuantile(record.cpu || 0)
    if (record.heapSizeLimit == null) {
      // default for 64bit and default flags
      record.heapSizeLimit = 1501560832
    }
    const memory = this.toQuantile((record.heapUsed || 0) / record.heapSizeLimit)

    let entry = this.cache.entries().find((e) => {
      return e.time === timeBucket
    })
    if (entry == null) {
      entry = new Entry(timeBucket)
      this.cache.insert(entry)
    }

    const coords = `${cpu},${memory}`
    if (entry.gridspace[coords] == null) {
      entry.gridspace[coords] = 1
    } else {
      entry.gridspace[coords]++
    }
  }

  backfill () {
    const entries = this.cache.entries()

    if (entries.length < 1) {
      this.seed(Date.now())
    } else {
      const entry = entries[entries.length - 1]
      if (entry && entry.time) {
        this.seed(entry.time)
      } else {
        this.seed(Date.now())
      }
    }
  }

  entries () {
    // backfill any empty time slots
    this.backfill()
    const entries = this.cache.entries()
    if (entries.length > BUCKETS) {
      return entries.slice(entries.length - BUCKETS, entries.length)
    }
    return entries.slice()
  }
}

module.exports = QuantileHistory
