// console server
'use strict'

var url = require('url')
var path = require('path')
var spawn = require('child_process').spawn
var split = require('split')
var createMux = require('multiplex')

const bole = require('bole')
const logger = bole(process.title)('server')

var createServerRoutes = require('./routes')
var storage = require('./storage')
var json = require('./json')

var st = require('st')
var skateboard = require('skateboard')

var defaultNodeEnv = 'production'

function bailOnError (message, fn, context) {
  return function () {
    logger.error(message)
    try {
      return fn.apply(context, arguments)
    } catch (e) {
      logger.error(e)
    }
  }
}

function createServer (options, pollingEmitter) {
  var developmentMode = (process.env.NODE_ENV || defaultNodeEnv) !== 'production'
  var workerPath = path.join(__dirname, 'heap-snapshot', 'worker.js')
  var applicationClients = new Set()

  pollingEmitter.on('output', function (output) {
    const payload = Object.assign({
      serverTime: Date.now()
    }, output)
    broadcastToApplicationClients(payload)
  })

  function removeConnectedStream (stream) {
    if (applicationClients.has(stream)) {
      logger.debug('Removing closed client stream')
      applicationClients.delete(stream)
    }
    stream.destroy()
  }

  function setupNewMetricsStream (stream) {
    // keep track of connected clients
    applicationClients.add(stream)
    logger.info('Adding new client stream (%s connections)', applicationClients.size)

    var metricsStream
    var vulnerabilitiesStream
    var assetsStream
    var thresholdsStream

    const cleanup = () => {
      removeConnectedStream(stream)
      if (metricsStream) {
        logger.debug('destroy existing metricsStream stream')
        metricsStream.destroy()
        metricsStream = null
      }

      if (assetsStream) {
        logger.debug('destroy existing assetsStream stream')
        assetsStream.destroy()
        assetsStream = null
      }

      if (vulnerabilitiesStream) {
        logger.debug('destroy existing vulnerabilitiesStream stream')
        vulnerabilitiesStream.destroy()
        vulnerabilitiesStream = null
      }

      if (thresholdsStream) {
        logger.debug('destroy existing thresholdsStream stream')
        thresholdsStream.destroy()
        thresholdsStream = null
      }
    }

    stream.on('end', cleanup)
    stream.on('error', cleanup)
    stream.on('close', cleanup)

    const initialPayload = Object.assign({}, {
      serverTime: Date.now()
    })

    const assets = {
      snapshot: new Map(),
      profile: new Map()
    }

    vulnerabilitiesStream = storage.createVulnerabilitiesStream(options)
    vulnerabilitiesStream.on('data', function (vulnerabilitiesPayload) {
      pollingEmitter.emit('output', vulnerabilitiesPayload)
    })

    let processCache

    stream.pipe(split()).on('data', function (line) {
      const obj = json(line)

      if (metricsStream) {
        logger.debug('destroy existing metricsStream stream')
        metricsStream.destroy()
      }

      if (assetsStream) {
        logger.debug('destroy existing assetsStream stream')
        assetsStream.destroy()
      }

      if (thresholdsStream) {
        logger.debug('destroy existing thresholdsStream stream')
        thresholdsStream.destroy()
      }

      if (!obj || !obj.app) {
        return
      }

      metricsStream = storage.createApplicationStream(options, {
        app: obj.app
      })

      metricsStream.on('data', function (applicationPayload) {
        // TODO-2.x: bombproofing
        const incoming = applicationPayload || { processes: [] }
        processCache = new Set(incoming.processes)

        pollingEmitter.emit('output', {
          application: {
            name: incoming.app,
            processes: incoming.processes
          }
          // TODO-2.x: make this request out of band
          // hosts: hosts
        })
      })

      assetsStream = storage.createAssetsStream(options, {
        app: obj.app
      })

      assetsStream.on('data', function (assetsPayload) {
        if (!assets[assetsPayload.type]) {
          logger.warn('no assets map for', assetsPayload.type)
          return
        }

        let aliveTmp = false
        if (processCache && processCache.size > 0) {
          processCache.forEach(process => {
            if (process.id === assetsPayload.id) { aliveTmp = true }
          })
        }
        assetsPayload.alive = aliveTmp

        assets[assetsPayload.type].set(assetsPayload.asset, assetsPayload)

        const output = {
          profiles: {
            [obj.app]: Array.from(assets.profile.values())
          },
          snapshots: {
            [obj.app]: Array.from(assets.snapshot.values())
          }
        }
        pollingEmitter.emit('output', output)
      })

      thresholdsStream = storage.createThresholdsStream(options, {
        start: '-' + Date.now()
      })

      thresholdsStream.on('data', function (thresholds) {
        pollingEmitter.emit('output', thresholds)
      })
    })

    stream.write(JSON.stringify(initialPayload) + '\n')
  }

  function setupNewGraphsStream (stream) {
    var graphStreams = new Set()
    function cleanup () {
      for (let graphStream of graphStreams) {
        if (graphStream.response) {
          graphStream.response.socket.end()
        } else {
          graphStream.on('error', (e) => {
            logger.error(e, 'DESTROY ERROR')
          })
          graphStream.destroy()
        }
        graphStreams.delete(graphStream)
      }
    }

    stream.on('end', cleanup)
    stream.on('error', cleanup)
    stream.on('close', cleanup)

    var bulk = ''
    var corkedLength = 0
    var corkTimer = null
    const TIMEOUT = 10
    const THRESHOLD = 65536

    function write (str) {
      if (!str.trim()) {
        return
      }

      clearTimeout(corkTimer)

      bulk += str
      corkedLength += Buffer.byteLength(str)

      if (corkedLength > THRESHOLD) {
        uncork()
      } else {
        corkTimer = setTimeout(uncork, TIMEOUT)
      }
    }

    function uncork () {
      stream.write(bulk)
      bulk = ''
      corkedLength = 0
    }

    // TODO-2.x: handle graph subscriptions
    stream.graphSubscriptions = new Map()
    stream.pipe(split()).on('data', function (line) {
      logger.debug(`Received graph subscription request ${line}`)

      if (!line.trim().length) {
        return
      }

      const subscription = json(line)
      if (!subscription) {
        return
      }

      logger.debug(`Got a subscription request for process ${subscription.processId}`)

      const subName = `process:${subscription.processId}`
      if (stream.graphSubscriptions.has(subName)) {
        logger.debug(`Already subscribed to ${subscription.processId}`)
        return
      }

      const activeSubscriptions = stream.graphSubscriptions
      const activeSubscriptionKeys = Array.from(activeSubscriptions.keys())
      if (activeSubscriptionKeys.length) {
        activeSubscriptionKeys.forEach(function (graphName) {
          const graphStream = activeSubscriptions.get(graphName)
          activeSubscriptions.delete(graphName)

          if (graphStream) {
            graphStream.on('error', (e) => {
              logger.error(e, 'DESTROY ERROR')
            })
            return graphStream.destroy()
          }
        })
      }

      // Superset of all the graphed fields
      const graphFields = [
        'activeHandles', 'activeRequests', 'cpu',
        'heapTotal', 'heapUsed', 'load1m',
        'percentCpuThreshold', 'percentHeapThreshold',
        'percentHostMemoryUsed', 'rss'
      ]
      var graphStream = storage.createMetricStream(options, {
        start: '-15m',
        field: graphFields,
        id: subscription.processId
      })

      graphStream.on('data', (obj) => {
        // Mutate the values to what the graphs expect
        if (obj.cpu) obj.cpu = obj.cpu * 100
        if (obj.percentHostMemoryUsed) obj.percentHostMemoryUsed = obj.percentHostMemoryUsed * 100
        if (obj.percentCpuThreshold === -1) obj.percentCpuThreshold = null

        write(JSON.stringify(obj) + '\n')
      })
      graphStreams.add(graphStream)

      stream.graphSubscriptions.set(subName, graphStream)

      logger.debug('active subscriptions', activeSubscriptions.keys())
    })
  }

  function setupRpcChannel (stream) {
    stream.pipe(split()).on('data', (message) => {
      var msg = json(message)
      if (msg && msg.command == null) {
        // ignoring malformed or empty content
        return
      }
      if (msg.command === 'summaryMetrics' && msg.app) {
        const summary = storage.getSummaryData(msg.app)
        if (summary == null) { return }
        summary.command = msg.command
        stream.write(JSON.stringify(summary) + '\n')
      } else if (msg.command === 'spiffy' && msg.app) {
        // TODO
      } else {
        // TODO move to debug
        logger.info('unable to decode RPC channel message', message)
      }
    })
  }

  function broadcastToApplicationClients (object) {
    const payload = JSON.stringify(object)
    for (let stream of applicationClients) {
      // TODO-2.x: improve reconnection
      stream.write(payload + '\n')
    }
  }

  var router = createServerRoutes(options, broadcastToApplicationClients)

  function requestHandler (req, res) {
    const urlPath = url.parse(req.url).pathname

    const match = router && router.match(urlPath)
    if (match) {
      // serve dynamic routes such as profiles and heap snapshots
      match.fn(req, res, match)
    } else {
      // if we got here, the user is probably following a deep link
      // send them to `/` so that the application code goes over the wire
      // the client will take care of doing the right thing with the URL
      req.url = '/'
      req.sturl = '/'
      staticHandler(req, res)
    }
  }

  const staticHandler = st({
    path: path.join(__dirname, '..', '..', 'build'),
    index: 'index.html',
    passthrough: true,
    cache: !developmentMode
  })

  function connectionHandler (stream, params) {
    if (params.source === 'applications') {
      var mux = createMux(function (incomingStream, id) {
        switch (id) {
          case 'metrics':
            setupNewMetricsStream(incomingStream)
            break

          case 'graphs':
            setupNewGraphsStream(incomingStream)
            break

          case 'rpc':
            setupRpcChannel(incomingStream) // TODO cleanup of this
            break

          default:
            console.error('unhandled id in connectionHandler:', id)
            break
        }
      })

      stream.on('end', mux.destroy.bind(mux))
      stream.on('error', mux.destroy.bind(mux))
      stream.on('close', mux.destroy.bind(mux))

      stream.pipe(mux).pipe(stream)
    } else if (params.source === 'heapsnapshot') {
      const proc = spawn(process.execPath,
        [
          workerPath,
          '--storageUrl', options.storageUrl,
          '--storageAPIVersion', options.storageAPIVersion
        ], {
          stdio: 'pipe'
        })

      stream.once('end', bailOnError('client disconnected from heapsnapshot worker', proc.kill, proc))
      proc.on('exit', bailOnError('heapsnapshot worker exited', stream.close, stream))
      // bailOnError already called by above handlers so just log error message
      proc.on('error', err => logger.error(err, 'Error launching heapsnapshot worker'))
      proc.stderr.pipe(process.stderr)
      stream.pipe(proc.stdin)

      proc.stdout.setEncoding('utf8')
      proc.stdout.pipe(stream)
    }
  }

  logger.info(`starting server on http://${options.consoleHostname}:${options.consolePort}`)
  return skateboard({
    interface: options.consoleHostname,
    port: options.consolePort,
    dir: path.join(__dirname, '..', 'build'),
    requestHandler: requestHandler,
    staticHandler: staticHandler
  }, connectionHandler)
}

// INTERFACE
module.exports.createServer = createServer
