'use strict'

const lodash = require('lodash')
const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

exports.create = create
exports.calculatePHMU = calculatePHMU
exports.calculatePHMUByProcess = calculatePHMUByProcess

const thresholdActions = require('../threshold-actions')
const thresholdActionsAddProcess = thresholdActions.addProcess
const getThresholdConfig = thresholdActions.getThresholdConfig

let QueuedInfo = null

// Create and return a new N|Solid agent.
function create (context, id, info, metrics) {
  if (QueuedInfo == null) createQueuedInfoHandler(context)

  return new Agent(context, id, info, metrics)
}

function createQueuedInfoHandler (context) {
  QueuedInfo = []

  let interval = setInterval(writeQueuedInfo, 1000)
  interval.unref()

  // check to write queued info
  function writeQueuedInfo () {
    // Note that some info records may be written between when influx becomes
    // available, and the interval fires; resulting in out-of-order writes.
    // But shouldn't matter, as time stamps from the process start time.
    if (context.influxdbWriter == null) return

    // once influx is available, cancel the interval
    clearInterval(interval)

    for (let info of QueuedInfo) {
      const agent = info.agent
      const message = info.message

      Logger.debug(`writing queued info to influx: ${agent.id}: hostname: ${agent.tags.hostname}, app: ${agent.tags.app}`)
      context.influxdbWriter.writeInfo(agent, message)
    }

    QueuedInfo = []
  }
}

// Models an N|Solid agent.
class Agent {
  constructor (context, id, info, metrics) {
    this._context = context
    this.id = id
    this.tags = {}
    this.info = info // may be null
    this.metrics = metrics // may be null
    this._queuedMetrics = []
    this.packagesHash = null
    this.pendingCommandTimers = new Set()

    if (info) {
      this.addInfo(info)
    }

    if (metrics) {
      this.addMetrics(metrics)
    }

    Logger.debug(`creating new agent: ${id}`)
  }

  addPackages (packages) {
    if (!packages || !packages.length) {
      return
    }
    this.packagesHash = this._context.vulnManager.addPackages(packages)
    this.packages = packages
  }

  addInfo (message) {
    if (!message || !message.body) {
      return
    }

    this.info = lodash.clone(message.body)

    if (message.body.app) {
      this.tags.app = message.body.app
      delete message.body.app
    }

    if (message.body.hostname) {
      this.tags.hostname = message.body.hostname
      delete message.body.hostname
    }

    if (this._context.influxdbWriter == null) {
      Logger.debug(`queuing info to influx: ${this.id}: hostname: ${this.tags.hostname}, app: ${this.tags.app}`)
      QueuedInfo.push({agent: this, message: message})
    } else {
      Logger.debug(`adding info to agent: ${this.id}: hostname: ${this.tags.hostname}, app: ${this.tags.app}`)
      this._context.influxdbWriter.writeInfo(this, message)

      if (this._queuedMetrics.length) {
        this._queuedMetrics.forEach(metrics => {
          metrics.body = this.formatMetricsForWriting(metrics.body)
          this._context.influxdbWriter.writeMetrics(this, metrics)
        })
        this._queuedMetrics = []
      }
    }
  }

  addMetrics (message) {
    if (!message || !message.body) {
      return
    }

    if (this.info) {
      const thresholdConfig = getThresholdConfig(this._context)
      const conf = thresholdConfig[this.info.app]
      message.body.percentCpuThreshold = conf && conf.cpu && conf.cpu.enabled ? conf.cpu.threshold : -1
      message.body.percentHeapThreshold = conf && conf.heap && conf.heap.enabled ? conf.heap.threshold : -1
    }

    if (this.info && this._context.influxdbWriter) {
      this.metrics = message.body = this.formatMetricsForWriting(message.body)
      thresholdActionsAddProcess(message.body, this.info)
      this._context.influxdbWriter.writeMetrics(this, message)
    } else {
      this._queuedMetrics.push(message)
    }
  }

  // Used when writing metrics to the time-series database.
  formatMetricsForWriting (metrics) {
    // convert cpu % from 0..100 to 0..1 for consistency with other
    // percentile values
    if (metrics.cpu != null) {
      metrics.cpu = Math.round(10000 * metrics.cpu / 100) / 10000
    }

    return metrics
  }

  // Used when returning metrics from POST /metrics command.
  formatMetricsForReturning (metrics) {
    metrics = this.formatMetricsForWriting(metrics)
    metrics = this.addPercentMemoryProperties(metrics)

    if (this.info != null) {
      metrics.app = this.info.app
      metrics.hostname = this.info.hostname
      metrics.tags = this.info.tags
    }

    return metrics
  }

  // Used when returning metrics from GET /metrics command.
  formatMetricsFromQuery (metrics) {
    metrics = this.addPercentMemoryProperties(metrics)

    if (this.info != null) {
      metrics.app = this.info.app
      metrics.hostname = this.info.hostname
      metrics.tags = this.info.tags
    }

    return metrics
  }

  // Adds the calculated fields to metrics using info.
  addPercentMemoryProperties (metrics) {
    if (!metrics || !this.info) return metrics

    if (metrics.rss != null) {
      metrics.percentHostMemoryUsedByProcess = calculatePHMUByProcess({
        rss: metrics.rss,
        totalMem: this.info.totalMem
      })
    }

    if (metrics.freeMem != null) {
      metrics.percentHostMemoryUsed = calculatePHMU({
        totalMem: this.info.totalMem,
        freeMem: metrics.freeMem
      })
    }

    return metrics
  }

  getCacheRecord () {
    var obj = {
      id: this.id,
      app: this.tags.app,
      hostname: this.tags.hostname,
      tags: canonicalizeTags(this.info && this.info.tags),
      info: this.info,
      metrics: this.metrics
    }
    return obj
  }

  dispose () {
    this._context = null
    for (var timer of this.pendingCommandTimers) {
      clearTimeout(timer)
      this.pendingCommandTimers.delete(timer)
    }
  }
}

function calculatePHMUByProcess (params) {
  if (!params.rss || !params.totalMem) {
    return 0
  }
  return Math.round(10000 * params.rss / params.totalMem) / 10000
}

function calculatePHMU (params) {
  if (!params.freeMem || !params.totalMem) {
    return 0
  }
  return Math.round(10000 * (params.totalMem - params.freeMem) / params.totalMem) / 10000
}

// Normalize tags as null, an array, or string, to an array.
function canonicalizeTags (tags) {
  if (tags == null) return []
  if (Array.isArray(tags)) return tags

  // sometimes tags have leading and trailing commas ...
  return tags.split(/,/g).filter((tag) => tag !== '')
}
