'use strict'

const fs = require('fs')
const path = require('path')
const EventEmitter = require('events')

const async = require('async')
const mkdirp = require('mkdirp')

const util = require('./util')
const diskAssetGC = require('./disk-asset-gc')

const INITIAL_CLEANUP_SECONDS = 5

const Logger = util.getLogger(__dirname, __filename)

exports.create = create

function create (context) {
  return new DiskAssetManager(context)
}

// Manages assets on disk.
// Emits two events:
// * ("add", asset) - when an asset is added
// * ("remove", asset) - when the manager realizes an asset has been removed
class DiskAssetManager extends EventEmitter {
  // Create a new instance of this class, based on existing index file.
  constructor (context) {
    super()

    if (context.assetManager == null) {
      // dev-time error; assetManager must be set up before diskAssetManager
      throw new Error('assetManager not registered in context')
    }

    this._context = context
    this._assets = new Map()
    this._assetsDir = path.join(context.config.data.dir, 'assets')
    this._indexFile = path.join(this._assetsDir, 'index.json')
    this._isClosed = false
    this._writeQueue = async.queue(
      this._writeIndexToDiskTask.bind(this),
      1
    )

    const cleanupMinutes = context.config.assets.cleanupMinutes

    context.assetManager.on('add', (asset) => this.add(asset))

    this._readIndexFromDisk()

    this._cleanupInterval = setInterval(
      () => this.cleanUp(),
      cleanupMinutes * 60 * 1000
    )
    this._cleanupInterval.unref()

    setTimeout(
      () => this.cleanUp(),
      INITIAL_CLEANUP_SECONDS * 1000
    ).unref()
  }

  // Return an asset given it's assetID.
  byId (assetId) {
    return this._assets.get(assetId)
  }

  shutdown () {
    clearInterval(this._cleanupInterval)
    this._isClosed = true
  }

  // New asset being created, wait for complete.
  add (zmqAsset) {
    Logger.debug(`add(${zmqAsset.bulkId})`)
    zmqAsset.once('complete', (err) => this.complete(err, zmqAsset))
  }

  // Asset transfer complete, add to index.
  complete (err, zmqAsset) {
    if (err) {
      return Logger.error(err, 'error completing asset')
    }

    Logger.debug(`complete(${zmqAsset.bulkId})`)
    const relFileName = path.relative(this._assetsDir, zmqAsset.getFileName())
    const asset = {
      time: Date.now(),
      asset: zmqAsset.bulkId,
      type: zmqAsset.type,
      id: zmqAsset.agentId,
      app: 'unknown',
      hostname: 'unknown',
      tags: [],
      fileName: relFileName,
      size: zmqAsset.size,
      compressed: zmqAsset._compressed
    }

    const agentManager = this._context.agentManager
    const agent = agentManager.byId(asset.id)
    if (agent) {
      const cacheRecord = agent.getCacheRecord()

      if (cacheRecord.app) asset.app = cacheRecord.app
      if (cacheRecord.hostname) asset.hostname = cacheRecord.hostname
      if (cacheRecord.tags) asset.tags = cacheRecord.tags.slice()
      if (cacheRecord.info && cacheRecord.info.pid) asset.pid = cacheRecord.info.pid
      if (cacheRecord.metrics && cacheRecord.metrics.title) asset.title = cacheRecord.metrics.title
    }

    this._assets.set(asset.asset, asset)

    this._writeIndexToDisk((err) => {
      if (err) {
        if (this.isClosed) {
          // Nothing to do, probably data dir is missing now
          return
        }

        Logger.error(err, 'error writing index to disk')
        // don't return, emit event anyway, listener may be waiting
      }

      this.emit('add', asset)
    })
  }

  // Run a query over the assets, filtering on the usual filtering specs.
  queryMatchesAsset (qData, asset) {
    qData = qData || {}

    const tags = qData.tags || []

    if (!assetMatch(asset, 'id', qData.id)) return false
    if (!assetMatch(asset, 'app', qData.app)) return false
    if (!assetMatch(asset, 'hostname', qData.hostname)) return false
    if (!assetMatch(asset, 'type', qData.type)) return false

    // manual check for tags
    for (let tag of tags) {
      if (asset.tags.indexOf(tag) === -1) {
        return false
      }
    }

    // check times
    if (qData.start != null) {
      if (asset.time < qData.start) return false
    }

    if (qData.end != null) {
      if (asset.time > qData.end) return false
    }

    return true
  }

  query (qData) {
    const result = []

    for (let asset of this._assets.values()) {
      if (!this.queryMatchesAsset(qData, asset)) continue

      result.push(asset)
    }

    return result
  }

  // Cleanup assets based on count and size config values.
  cleanUp () {
    diskAssetGC.run(this._context, (err) => {
      if (err) {
        Logger.error(err, 'error cleaning up assets')
        // don't return, run validate as well ...
      }

      this.validate()
    })
  }

  // Validate that all the assets known actually exist, rewrite index if some
  // are missing.
  validate () {
    Logger.debug('_validate()')
    const self = this
    let changed = false

    // validate a single asset
    function validateAsset (asset, cb) {
      self._validateAsset(asset, (err, changed_) => {
        if (err) return cb(err)

        changed = changed || changed_
        cb()
      })
    }

    // validate all the assets in parallel
    const assetsArr = Array.from(this._assets.values())
    async.map(assetsArr, validateAsset, (err) => {
      if (err) {
        Logger.error(err, 'error validating assets')
        return
      }

      if (changed) {
        self._writeIndexToDisk((err) => {
          if (err) return
          Logger.debug('updated assets index with deletions')
        })
      }
    })
  }

  // Read the asset index file, set internal assets map.
  // Synchronous, only used at startup.
  _readIndexFromDisk () {
    // read the index file
    let jsonContents
    try {
      jsonContents = fs.readFileSync(this._indexFile, 'utf8')
    } catch (err) {
      if (err.code !== 'ENOENT') { // if file doesn't exist, that's fine
        Logger.error(err, `error reading assets index ${this._indexFile}`)
      }
      return
    }

    // parse the index file contents
    let contents
    try {
      contents = JSON.parse(jsonContents)
    } catch (err) {
      Logger.error(err, `unable to parse JSON in ${this._indexFile}`)
      return
    }

    // rebuild the map based on the contents of the index file
    this._assets.clear()
    for (let assetID in contents) {
      this._assets.set(assetID, contents[assetID])
    }
  }

  // Return the assets map as a JS object.
  _assetsAsObject () {
    const contents = {}

    for (let assetID of this._assets.keys()) {
      contents[assetID] = this._assets.get(assetID)
    }

    return contents
  }

  // Queues up writing a new asset index file, from the internal assets map.
  _writeIndexToDisk (cb) {
    Logger.debug('_writeIndexToDisk()')
    // post a task to the queue, which will call _writeIndexToDiskTask({}, cb)
    this._writeQueue.push({}, cb)
  }

  // Write a new asset index file, from the internal assets map.
  _writeIndexToDiskTask (taskIsEmpty, cb) {
    Logger.debug('_writeIndexToDiskTask()')
    const self = this

    function mkdirpTask (cb) {
      Logger.debug('_writeIndexToDiskTask() mkdirpTask')
      mkdirp(self._assetsDir, (err, data) => {
        if (err && err.code !== 'EEXIST') {
          Logger.error(err, `unable to create assets directory ${self._assetsDir}`)
          return cb(err)
        }
        cb()
      })
    }

    // Write the file.
    function writeFileTask (cb) {
      Logger.debug('_writeIndexToDiskTask() writeFileTask')

      // create JSON from JS object
      const contents = self._assetsAsObject()
      const jsonContents = JSON.stringify(contents, null, 2)

      fs.writeFile(self._indexFile, jsonContents, (err) => {
        if (err && !self._isClosed) {
          Logger.error(err, `error writing assets index ${self._indexFile}`)
          return cb(err)
        }
        cb()
      })
    }

    // mkdirp, then write file
    async.series([
      mkdirpTask,
      writeFileTask
    ], cb)
  }

  // Validate that the asset known actually exists, return indication of whether
  // file was removed; true if it was, false otherwise.
  _validateAsset (asset, cb) {
    Logger.debug(`_validateAsset(${asset.asset})`)
    const assetFile = path.join(this._assetsDir, asset.fileName + (asset.compressed ? '.gz' : ''))

    fs.stat(assetFile, (err, stats) => {
      if (!err && stats.isFile()) return cb(null, false)

      this._assets.delete(asset.asset)
      Logger.debug(`asset was deleted ${asset.asset}`)
      setImmediate(() => this.emit('remove', asset))

      cb(null, true)
    })
  }
}

// Check asset's specified property for the specified value.
function assetMatch (asset, property, value) {
  // if they didn't specify a matching value, it matches!
  if (value == null) return true

  if (asset[property] === value) return true

  return false
}
