'use strict'

exports.handleRequest = handleRequest

const path = require('path')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)
const zlib = require('zlib')
const fs = require('fs')

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsStorage.context
  const config = context.config
  const qData = req.nsStorage.qData
  const diskAssetManager = context.diskAssetManager

  const assetPath = path.join(config.data.dir, 'assets')

  const asset = diskAssetManager.byId(qData.asset)
  if (asset == null) {
    res.status(404).send('asset not found')
    return
  }

  const fileName = path.join(assetPath, asset.fileName)

  Logger.debug(`sending asset ${qData.asset}`)

  const fileZipped = asset.compressed
  const acceptsGzip = req.acceptsEncodings('gzip')
  const headers = {
    'Content-Disposition': 'attachment; filename=' + path.basename(fileName)
  }
  var realFileName
  if (fileZipped) {
    realFileName = fileName + '.gz'
    fs.access(realFileName, (fs.constants || fs).R_OK, err => {
      if (err) {
        Logger.error(err, `Requested asset ${realFileName} not readable`)
        try {
          res.status(404).send('asset not found')
        } catch (err) {
          Logger.debug(err, 'error during cancelled/errored sendFile')
        }
        return
      }

      if (acceptsGzip) {
        headers['Content-Encoding'] = 'gzip'
      } else {
        // This is something else entirely.  We need to stream the file
        // through zlib on the way out.
        const unzipStream = zlib.createGunzip()
        var sentHeaders = false
        unzipStream.on('error', err => {
          Logger.error(err, `Error unzipping ${realFileName}`)
          if (sentHeaders) {
            res.end()
          } else {
            res.status(500).send('internal error')
            sentHeaders = true
          }
        })
        const fileStream = fs.createReadStream(realFileName)
        fileStream.on('error', err => {
          Logger.error(err, `error reading file ${realFileName}`)
          if (sentHeaders) {
            res.end()
          } else {
            res.status(500).send('internal error')
            sentHeaders = true
          }
        })
        fileStream.pipe(unzipStream)

        if (!sentHeaders) {
          res.set(headers)
          sentHeaders = true
        }
        res.on('error', err => {
          Logger.error(err, `error sending file ${realFileName}`)
        })
        unzipStream.pipe(res)
      }
    })
    return
  } else {
    realFileName = fileName
  }

  res.sendFile(realFileName, { headers }, (err) => {
    if (err) {
      // write error if wasn't client closing the socket
      if (['EPIPE', 'ECONNABORTED'].indexOf(err.code) === -1) {
        Logger.error(err, 'error sending file')
      }

      // At this point, status/headers may have been sent, if response
      // is closed in the middle by the client, so we wrap this lest it throw.
      try {
        res.status(404).send('asset not found')
      } catch (err) {
        Logger.debug(err, 'error during cancelled/errored sendFile')
      }
    }
  })
}
