'use strict'

exports.handleRequest = handleRequest

const lodash = require('lodash')

const util = require('../util')

const Logger = util.getLogger(__dirname, __filename)

const AllFields = [
  'app', 'pid', 'execPath', 'main', 'tags', 'versions', 'cpuCores', 'cpuModel',
  'arch', 'platform', 'hostname', 'totalMem', 'processStart'
]
exports.allFields = AllFields

// Cause "new" data to be delayed by this duration before it is queried for
// polling requests.  It should be enough buffer that both agents and Influx
// have time to get the data stored before the window is queried.
const RECENT_DELAY = 1000

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsStorage.context
  const eventStream = req.nsStorage.eventStream
  const qData = req.nsStorage.qData
  const readAtInterval = qData.id == null && qData.end == null

  let contInterval = null
  if (readAtInterval) {
    contInterval = setInterval(readNextMetrics, 1000)
    contInterval.unref()
  }

  // Normalize the query data.
  if (qData.field == null || qData.field.length === 0) qData.field = AllFields

  if (qData.id == null) {
    if (qData.start == null) qData.start = Date.now() - 60 * 1000
    if (qData.end == null) qData.end = Date.now() - RECENT_DELAY
  } else {
    qData.start = null
    qData.end = null
  }

  let diff = lodash.difference(qData.field, AllFields)
  for (let field of diff) {
    Logger.warn(`for ${JSON.stringify(req.query)}: ignoring invalid field ${field}`)
  }

  qData.field = lodash.intersection(qData.field, AllFields)
  qData.field.unshift('id')

  const query = buildQuery(qData)

  Logger.debug(`running query: ${query}`)
  readMetrics(context, query, readMetricsDone)

  // read metrics, returning empty data if influx hasn't started yet
  function readMetrics (context, query, cb) {
    if (context.influxdbReader == null) {
      setImmediate(cb, null, '{}')
    } else {
      context.influxdbReader.readMetrics(query, cb)
    }
  }

  function readMetricsDone (err, data) {
    if (err) {
      Logger.error(err, 'error reading metrics')

      eventStream.close()
      return
    }

    if (eventStream.isClosed()) return

    const events = eventsFromQuery(data)
    for (let event of events) {
      eventStream.writeData(event)
    }

    if (contInterval == null) {
      eventStream.close()
    }
  }

  function readNextMetrics () {
    if (eventStream != null && eventStream.isClosed()) {
      clearInterval(contInterval)
      contInterval = null
      return
    }

    qData.start = qData.end
    qData.end = Date.now() - RECENT_DELAY

    const query = buildQuery(qData)

    Logger.debug(`running query: ${query}`)
    readMetrics(context, query, readMetricsDone)
  }
}

// Build the query from the query string data.
function buildQuery (qData) {
  const queryParts = ['SELECT']

  const queryFields = qData.field.map(f => fieldRef(f, qData.agg))

  queryParts.push(queryFields)

  queryParts.push('FROM info WHERE')

  const whereParts = []

  if (qData.start != null) {
    whereParts.push(`time >= '${util.dateMS2ISO(qData.start)}'`)
    whereParts.push(`time < '${util.dateMS2ISO(qData.end)}'`)
  }

  if (qData.id) {
    whereParts.push(`id = '${qData.id}'`)
  } else {
    if (qData.app) whereParts.push(`app = '${qData.app}'`)
    if (qData.hostname) whereParts.push(`hostname = '${qData.hostname}'`)
  }

  queryParts.push(whereParts.join(' AND '))

  return queryParts.join(' ')
}

// Return a field reference given the field name and aggregation
function fieldRef (field) {
  return `"${field}"`
}

// Create events from the query results.
function eventsFromQuery (data) {
  try {
    data = JSON.parse(data)
  } catch (err) {
    Logger.warn(err, 'error parsing JSON data from query', data)
    return []
  }

  if (data.results == null) return []
  if (data.results.length === 0) return []
  if (data.results[0].series == null) return []
  if (data.results[0].series.length === 0) return []

  const series = data.results[0].series[0]
  const events = []

  for (let i = 0; i < series.values.length; i++) {
    // let skip = false
    // for (let j = 0; j < series.columns.length; j++) {
    //   if (series.values[i][j] == null) skip = true
    // }
    // if (skip) continue

    var event = {}
    for (let j = 0; j < series.columns.length; j++) {
      event[series.columns[j]] = series.values[i][j]
    }

    if (event.versions) {
      try {
        event.versions = JSON.parse(event.versions)
      } catch (err) {
        Logger.error('Can not parse info.versions', err)
      }
    }

    events.push(event)
  }

  return events
}
