'use strict'

const lodash = require('lodash')

const api = require('./api')
const util = require('../util')

exports.create = create

// Buffer writes for this many millis.
const BUFFER_FOR_MS = 250
const CHUNK_SIZE = 100

const Logger = util.getLogger(__dirname, __filename)

// Create and return a new database writer.
function create (context) {
  return new Writer(context)
}

class Writer {
  constructor (context) {
    this.context = context
    this.port = this.context.config.influxdb.portHTTP
    this.timeout = null
    this.queue = []
    this.writing = false
  }

  // Write an info record to the database.
  writeInfo (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)
    data.id = agent.id

    if (data.tags) {
      Logger.debug(`data.tags: ${JSON.stringify(data.tags)}`)
      if (lodash.isString(data.tags)) {
        data.tags = `,${data.tags},`
      } else if (lodash.isArray(data.tags)) {
        data.tags = `,${data.tags.join(',')},`
      } else {
        Logger.warn(`invalid type of tags: ${JSON.stringify(data.tags)}`)
        delete data.tags
      }
    }
    message.body = data

    // There should be a processStart field, so use that as the influx
    // time-stamp, which should be in ns instead of ms.
    if (data.processStart != null) {
      message.timeNS = `${data.processStart}000000` // big nums! don't use math!
    }

    const line = this._createLine(agent, 'info', message)
    this._write(line)
  }

  // Write an metrics record to the database.
  writeMetrics (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)
    data.id = agent.id

    message.body = data
    const line = this._createLine(agent, 'metrics', message)
    this._write(line)
  }

  // influx keys and tags must escape "=", " ", and ","
  escapeKey (key) {
    return key.replace(/([ =,])/g, '\\$1')
  }

  // format influx field values
  escapeFieldValue (value) {
    if (value == null) return '"null"'  // return null as the string 'null'
    if (lodash.isNumber(value)) return value
    if (lodash.isBoolean(value)) return `${value}`

    // if it smells like a string, ensure it's a string
    if (lodash.isString(value)) {
      value = `${value}`
    // last chance - it's an object, so JSON stringify
    } else {
      value = JSON.stringify(value)
    }

    // return quoted string, escaping quotes
    return `"${value.replace(/"/g, '\\"')}"`
  }

  _createLine (agent, measurement, message) {
    const tags = Object.keys(agent.tags).length
      ? ',' + lodash.map(agent.tags, (val, key) => `${this.escapeKey(key)}=${this.escapeKey(val)}`).join(',')
      : ''

    const fields = lodash.map(message.body, (val, key) => `${this.escapeKey(key)}=${this.escapeFieldValue(val)}`).join(',')

    const timestamp = message.timeNS || util.uniqDateNS()
    return `${measurement}${tags} ${fields} ${timestamp}`
  }

  // Write a line to the specified database.
  _write (line) {
    Logger.debug(`queuing line to InfluxDB: ${line}`)
    this.queue.push(line)

    if (this.queue.length >= CHUNK_SIZE) {
      this._writeQueued()
      return
    }

    if (this.timeout) return

    this.timeout = setTimeout(() => this._writeQueued(), BUFFER_FOR_MS)
    this.timeout.unref()
  }

  _writeQueued () {
    if (this.queue.length === 0) return

    // Only write one chunk at a time.  If an extended buffering period has
    // been in place, there could be many thousands of lines to process, so
    // don't stampede.
    if (this.writing) return
    this.writing = true

    Logger.debug(`writing ${this.queue.length} queued line to InfluxDB`)

    // Write the first 100 lines in the queue
    const joinedLines = this.queue.slice(0, CHUNK_SIZE).join('\n')
    this.queue = this.queue.slice(CHUNK_SIZE)

    api.write(this.port, 'nsolid', joinedLines, err => {
      this.writing = false
      if (err) {
        Logger.error(err, 'error while writing metrics to Influx, data discarded')
      }
      if (this.queue.length >= CHUNK_SIZE) {
        this._writeQueued()
      }
    })

    // Do this synchronously because data can arrive while influx is writing
    // and we want to allow a new timer to start in the meantime.
    if (this.timeout) clearTimeout(this.timeout)
    this.timeout = null
  }
}
