'use strict'

const fs = require('fs')
const path = require('path')

const lodash = require('lodash')

const util = require('./util')

const Logger = util.getLogger(__dirname, __filename)

exports.create = create

function create (context) {
  return new SettingsManager(context)
}

// Manages the application settings.
class SettingsManager {
  constructor (context) {
    this._context = context
    this._file = path.join(context.config.data.dir, 'settings.json')

    this._settings = readSettings(this._file)
  }

  // Return the current settings.
  getSettings () {
    return lodash.cloneDeep(this._settings)
  }

  // Set and persist the specified settings.
  setSettings (settings, cb) {
    writeSettings(this._file, settings, (err) => {
      if (err) return cb(err)

      this._settings = lodash.cloneDeep(settings)
      cb()
    })
  }
}

// Read the settings from the specified file.
function readSettings (fileName) {
  let jsonContents
  try {
    jsonContents = fs.readFileSync(fileName, 'utf8')
  } catch (err) {
    if (err.code !== 'ENOENT') { // if file doesn't exist, that's fine
      Logger.error(err, `error reading settings file ${fileName}`)
    }
    return {}
  }

  // parse the index file contents
  let contents
  try {
    contents = JSON.parse(jsonContents)
  } catch (err) {
    Logger.error(err, `unable to parse JSON in ${fileName}`)
    return {}
  }

  return contents
}

// Write the settings to the specified file.
function writeSettings (fileName, settings, cb) {
  let jsonContents

  try {
    jsonContents = JSON.stringify(settings, null, 2)
  } catch (err) {
    const message = 'unable to stringify JSON'
    Logger.error(err, message)
    return cb(new Error(message))
  }

  fs.writeFile(fileName, jsonContents, (err) => {
    if (err) {
      const message = 'error writing settings file'
      Logger.error(err, message)
      return cb(new Error(message))
    }

    cb()
  })
}
