'use strict'

module.exports = profile

const hyperquest = require('hyperquest')
const url = require('url')
const concat = require('concat-stream')
const json = require('../json')
const buildUrlObj = require('../build-url-obj')
const profileCommand = require('../../http-api/POST-profile')

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// After the profile has reached its expected completion time, allow
// this duration before it is considered to be failed
const TIMEOUT_PAD_SECONDS = 3
profile.timeoutPadSeconds = TIMEOUT_PAD_SECONDS

function profile (appConfig, id, duration, fn) {
  // NB: if you're going to return early for debugging, be sure to call
  // fn({ app: 'app id', id: 'process id', asset: 'asset id' })
  // so that you don't hang the queue

  // currently the best we could do is to loop around and trigger these via the
  // HTTP api. In the long-term, we obviously need a lighter-weight, saner
  // solution that directly invokes the snapshot/profile functionality.
  // Getting there, however, will take a bit of internal refactoring
  const urlObj = buildUrlObj(appConfig.profileEndpoint.url, 'profile', appConfig, {
    id: id,
    duration: duration
  })

  // Don't call the async callback more than once, even if multiple errors occur
  fn = util.onlyCallOnce(fn)

  // Allow a few extra seconds for completion beyond the specified duration.
  // If it does not finish, possibly the agent died.
  const timer = setTimeout(() => {
    fn(new Error('Profile did not complete'))
  }, ((duration || profileCommand.defaultDuration) + profile.timeoutPadSeconds) * 1000)

  const profileUrl = url.format(urlObj)
  Logger.debug(`profile via ${profileUrl}`)

  hyperquest.post(profileUrl, (err, response) => {
    clearTimeout(timer)
    if (err) {
      return fn(err)
    }

    response.pipe(concat((payload) => {
      const obj = json(payload.toString())
      if (!obj) {
        return fn(new Error('Invalid response'))
      }

      fn(null, obj)
    }))
  })
  .on('error', res => {
    clearTimeout(timer)
    fn(new Error('Invalid response:', res))
  })
  .end()
}
