'use strict'

const EventEmitter = require('events').EventEmitter

class TimedCache extends EventEmitter {
  constructor (ttlMs, gcIntervalMs) {
    super()
    if (ttlMs >>> 0 !== ttlMs) {
      throw new Error('TTL should be an integer number of milliseconds')
    }
    this._ttlMs = ttlMs
    this._gcIntervalMs = gcIntervalMs || 1000
    this._cache = new Map()
    this._expiryLookup = new Map()
    this.timerLoop = setInterval(this.gc.bind(this), this._gcIntervalMs)
    this.timerLoop.unref()
  }

  purge () {
    this._cache = new Map()
    this._expiryLookup = new Map()
  }

  upsert (key, entry) {
    // add key, (re)set timer
    let added = !this._cache.has(key)
    this._cache.set(key, entry)
    this._expiryLookup.set(key, Date.now() + this._ttlMs)
    if (added) {
      this.emit('added', key, entry)
    }
  }

  get (key) {
    this.checkExpired(key)
    return this._cache.get(key)
  }

  delete (key) {
    if (!this._cache.has(key)) {
      return
    }
    var value = this._cache.get(key)
    this._delete(key)
    this.emit('deleted', key, value)
  }

  _delete (key) {
    this._cache.delete(key)
    this._expiryLookup.delete(key)
  }

  entries () {
    this.gc()
    var clone = {}
    for (var entry of this._cache) {
      clone[entry[0]] = entry[1]
    }
    return clone
  }

  checkExpired (key) {
    if (Date.now() > this._expiryLookup.get(key)) {
      this.emit('expired', key, this._cache.get(key))
      this._delete(key)
    }
  }

  gc () {
    this._expiryLookup.forEach((key, val) => {
      this.checkExpired(val)
    })
  }

  stopTimerLoop () {
    clearInterval(this.timerLoop)
  }
}

module.exports = TimedCache
