'use strict'

const zmqzap = require('zmq-zap')
const zap = new zmqzap.ZAP()
const util = require('../util')
const ZMQBindSocket = require('../zmqserver/socket')
const ZMQRemoteSocket = require('./remote-socket')
const BufferQueue = require('./buffer-queue')
const SharedMemoryManager = require('./shared-memory-manager')
const websocket = require('./websocket')

const Logger = util.getLogger(__dirname, __filename)

zap.use(new zmqzap.CurveMechanism(function (data, callback) {
  // Authorize all ZeroMQ socket connections.
  callback(null, true)
}))

class ZeroMQRelay {
  constructor (config) {
    this.config = config

    this.memoryManager = new SharedMemoryManager(config.relay)

    const zmqConfigured = (config.sockets.commandRemoteAddr &&
                           config.sockets.dataRemoteAddr &&
                           config.sockets.bulkRemoteAddr)
    const wsConfigured = !!config.sockets.wsRemoteUrl
    if (!(zmqConfigured || wsConfigured) ||
         (zmqConfigured && wsConfigured)) {
      throw new Error('Relay requires one of ZMQ or Websocket remotes be configured')
    }

    // data
    if (config.sockets.dataBindAddr) {
      this.dataPull = new ZMQBindSocket({
        bindUrl: config.sockets.dataBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.HWM,
        type: 'pull'
      })

      // Relay all `data` messages, queue if necessary.
      this.dataPull.socket.on('message', message => {
        this.handleData(message)
      })
    }

    if (config.sockets.dataRemoteAddr) {
      this.dataPush = new ZMQRemoteSocket({
        remoteUrl: config.sockets.dataRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.HWM,
        type: 'push'
      })
    }

    // create an engine.io listener
    if (this.config.sockets.wsBindAddr) {
      this.wsListener = new websocket.Listener(this)
    }

    // create an engine.io forwarder
    if (this.config.sockets.wsRemoteUrl) {
      this.wsForwarder = new websocket.Forward(this)
    }

    const dataBufferQueue = new BufferQueue({
      name: 'dataPush',
      config: config.relay,
      sendBuffered: buffers => {
        if (this.wsForwarder) {
          this.wsForwarder.sendData(buffers[0])
        }
        if (this.dataPush) {
          this.dataPush.socket.send(buffers)
        }
      }
    })
    this.dataBufferQueue = dataBufferQueue
    if (this.dataPush) {
      this.dataPush.socket.on('connect', () => {
        this.dataBufferQueue.resume()
      })
      this.dataPush.socket.on('disconnect', () => {
        this.dataBufferQueue.pause()
      })
    }
    if (this.wsForwarder) {
      this.wsForwarder.on('open', () => {
        this.dataBufferQueue.resume()
      })
      this.wsForwarder.on('close', () => {
        this.dataBufferQueue.pause()
      })
    }

    // Register the Data Queue in the memory manager
    this.memoryManager.registerBufferQueue(dataBufferQueue)

    // bulk
    if (config.sockets.bulkBindAddr) {
      this.bulkPull = new ZMQBindSocket({
        bindUrl: config.sockets.bulkBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.bulkHWM,
        type: 'pull'
      })

      // Relay all `bulk` messages, queue if necessary.
      this.bulkPull.socket.on('message', (message, data) => {
        this.handleBulk(message, data)
      })
    }

    if (config.sockets.bulkRemoteAddr) {
      this.bulkPush = new ZMQRemoteSocket({
        remoteUrl: config.sockets.bulkRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.bulkHWM,
        type: 'push'
      })
    }

    const bulkBufferQueue = new BufferQueue({
      name: 'bulkPush',
      config: config.relay,
      sendBuffered: buffers => {
        if (this.wsForwarder) {
          this.wsForwarder.sendBulk(buffers[0], buffers[1])
        }
        if (this.bulkPush) {
          this.bulkPush.socket.send(buffers)
        }
      }
    })
    this.bulkBufferQueue = bulkBufferQueue
    if (this.bulkPush) {
      this.bulkPush.socket.on('connect', () => {
        this.bulkBufferQueue.resume()
      })
      this.bulkPush.socket.on('disconnect', () => {
        this.bulkBufferQueue.pause()
      })
    }
    if (this.wsForwarder) {
      this.wsForwarder.on('open', () => {
        this.bulkBufferQueue.resume()
      })
      this.wsForwarder.on('close', () => {
        this.bulkBufferQueue.pause()
      })
    }

    // Register the Bulk Queue in the memory manager
    this.memoryManager.registerBufferQueue(bulkBufferQueue)

    // command
    if (config.sockets.commandBindAddr) {
      this.commandPub = new ZMQBindSocket({
        bindUrl: config.sockets.commandBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.HWM,
        type: 'xpub'
      })
      this.commandPub.socket.on('message', message => {
        // An XPUB message is a subscription.  The first byte in the buffer will be
        // 1 for subscribe and 0 for unsubscribe.  The remaining bytes are the
        // subscription string.
        var isSub = message[0]
        var subscription = message.slice(1).toString()
        if (isSub && /^configuration/.test(subscription)) {
          this.commandPub.socket.send([subscription, JSON.stringify({ config })])
        }
      })
    }

    if (config.sockets.commandRemoteAddr) {
      this.commandSub = new ZMQRemoteSocket({
        remoteUrl: config.sockets.commandRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.HWM,
        type: 'sub'
      })

      // Relay all `command` messages, no buffering or HWM
      this.commandSub.socket.on('message', (topic, message) => {
        this.handleCommand(topic, message)
      })
    }
  }

  logCommand (topic, message) {
    Logger.debug('Relaying a command message on %s -', topic.toString('utf8'), message.toString('utf8'))
  }

  handleCommand (topic, message) {
    if (Logger.getLogLevel() === 'debug') {
      this.logCommand(topic, message)
    }

    if (this.wsListener) {
      this.wsListener.sendCommand(topic, message)
    }
    if (this.commandPub) {
      this.commandPub.socket.send([topic, message])
    }
  }

  logData (msgBuffer) {
    const message = msgBuffer.toString('utf8')
    let parsed
    try {
      parsed = JSON.parse(message)
    } catch (e) {
      Logger.debug('Received a corrupted data message', message)
      return
    }

    if (parsed.command === 'metrics') {
      return // Skip the flood of metrics data
    }

    const status = this.dataBufferQueue.buffering ? 'Buffering' : 'Relaying'
    Logger.debug(`${status} a data message`, parsed.command)
  }

  handleData (message) {
    if (Logger.getLogLevel() === 'debug') {
      this.logData(message)
    }

    if (!this.dataBufferQueue.buffering) {
      if (this.wsForwarder) {
        this.wsForwarder.sendData(message)
      }
      if (this.dataPush) {
        this.dataPush.socket.send(message)
      }
      return
    }

    this.dataBufferQueue.queueBuffers([message])
  }

  logBulk (msgBuffer) {
    const message = msgBuffer.toString('utf8')
    let parsed
    try {
      parsed = JSON.parse(message)
    } catch (e) {
      Logger.debug('Received a corrupted bulk message', message)
      return
    }
    const status = this.bulkBufferQueue.buffering ? 'Buffering' : 'Relaying'
    Logger.debug(`${status} a bulk message`, parsed.command)
  }

  handleBulk (message, data) {
    if (Logger.getLogLevel() === 'debug') {
      this.logBulk(message)
    }

    if (!this.bulkBufferQueue.buffering) {
      if (this.wsForwarder) {
        this.wsForwarder.sendBulk(message, data)
      }
      if (this.bulkPush) {
        this.bulkPush.socket.send([message, data])
      }
      return
    }

    this.bulkBufferQueue.queueBuffers([message, data])
  }

  reloadConfig (config) {
    Logger.info('Reloading socket config')
    this.config = config

    // Only reload upstream sockets
    if (this.dataPush) {
      this.dataPush.setOptions(config.sockets.dataRemoteAddr, config.sockets.publicKey)
    }
    if (this.bulkPush) {
      this.bulkPush.setOptions(config.sockets.bulkRemoteAddr, config.sockets.publicKey)
    }
    if (this.commandSub) {
      this.commandSub.setOptions(config.sockets.commandRemoteAddr, config.sockets.publicKey)
    }
    if (this.wsForwarder) {
      this.wsForwarder.connect()
    }

    Logger.info('Reloading BufferQueue config')
    this.dataBufferQueue.setOptions(config.relay)
    this.bulkBufferQueue.setOptions(config.relay)
    this.memoryManager.setOptions(config.relay)
  }

  shutdown () {
    Logger.info('Shutting down ZeroMQ')
    if (this.dataPull) this.dataPull.close()
    if (this.dataPush) this.dataPush.close()
    if (this.bulkPull) this.bulkPull.close()
    if (this.bulkPush) this.bulkPush.close()
    if (this.commandSub) this.commandSub.close()
    if (this.commandPub) this.commandPub.close()
    if (this.wsForwarder) this.wsForwarder.close()
    if (this.wsListener) this.wsListener.close()

    Logger.info('Closing BufferQueues')
    this.bulkBufferQueue.close()
    this.dataBufferQueue.close()
  }
}

module.exports = ZeroMQRelay
