'use strict'

const zmqzap = require('zmq-zap')
const zap = new zmqzap.ZAP()
const util = require('../util')
const ZMQBindSocket = require('./socket')

const Logger = util.getLogger(__dirname, __filename)

zap.use(new zmqzap.CurveMechanism(function (data, callback) {
  // Authorize all ZeroMQ socket connections.
  callback(null, true)
}))

class ZeroMQServer {
  constructor (params) {
    this.config = params.config
    this.bindCb = util.onlyCallOnce(params.bindCb)

    var boundCount = 0
    var socksBound = err => {
      if (!this.bindCb) return
      if (err) this.bindCb(err)
      if (++boundCount === 3) this.bindCb()
    }

    this.data = new ZMQBindSocket({
      bindUrl: this.config.sockets.dataBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.HWM,
      type: 'pull',
      bindCb: socksBound
    })

    this.bulk = new ZMQBindSocket({
      bindUrl: this.config.sockets.bulkBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.bulkHWM,
      type: 'pull',
      bindCb: socksBound
    })

    this.command = new ZMQBindSocket({
      bindUrl: this.config.sockets.commandBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.HWM,
      type: 'xpub',
      bindCb: socksBound
    })
  }

  send (message) {
    if (message.id) {
      this.sendToAgent(message)
    } else {
      this.broadcast(message)
    }
  }

  sendToAgent (message) {
    if (!this.command.socket) {
      Logger.info('Message could not be sent after socket closed', message)
    }

    Logger.debug('sending agent command message', message)
    var id = message.id
    message = JSON.stringify(message)
    this.command.socket.send([`id:${id}`, message])
  }

  broadcast (message) {
    if (!this.command.socket) {
      Logger.info('Message could not be sent after socket closed', message)
    }

    Logger.debug('sending broadcast command message', message)
    message = JSON.stringify(message)
    this.command.socket.send(['broadcast', message])
  }

  shutdown () {
    Logger.info('Shutting down ZeroMQ')

    closeCatching('data', this.data)
    closeCatching('bulk', this.bulk)
    closeCatching('command', this.command)
  }
}

function closeCatching (name, socket) {
  try {
    socket.close()
  } catch (err) {
    Logger.warn(err, `error closing zmq socket ${name}`)
  }
}

module.exports = ZeroMQServer
